﻿using Dapper;
using EVCB_OCPP.TaskScheduler.Models;
using Quartz;
using System;
using System.Collections.Generic;
using System.Configuration;
using System.Data.SqlClient;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace EVCB_OCPP.TaskScheduler.Jobs
{
    /// <summary>
    /// Update  EVSE's online status
    /// </summary>
    [DisallowConcurrentExecution]
    public class CheckEVSEOnlineJob : IJob
    {
        private NLog.ILogger logger = NLog.LogManager.GetCurrentClassLogger();
        private DateTime latestHeartbeatTime = DateTime.Now;
        private List<EVSEOnlineRecord> updateData = new List<EVSEOnlineRecord>();
        private List<EVSEOnlineRecord> insertData = new List<EVSEOnlineRecord>();
        private string mainDBConnectString = ConfigurationManager.ConnectionStrings["MainDBContext"].ToString();
        private string onlineDBConnectString = ConfigurationManager.ConnectionStrings["OnlineLogDBContext"].ToString();

        public CheckEVSEOnlineJob()
        {
           
        }
       
        public async Task Execute(IJobExecutionContext context)
        {
            logger.Debug(this.ToString() + " :Starting........");
            try
            {
                List<EVSECurrentStatus> _EVSEs = GetEVSEs();
                var checktime = DateTime.UtcNow.AddDays(-3);
                _EVSEs = _EVSEs.Where(x => x.HeartbeatUpdatedOn > checktime).ToList();

                foreach (var evse in _EVSEs)
                {
                    latestHeartbeatTime = evse.HeartbeatUpdatedOn;
                    if (IsOnlineNow(evse))
                    {

                        if (!evse.Online)
                        { //off - on 
                            logger.Debug(evse.ChargeBoxId + " Off-On Started");

                            UpdateEVSECurrentStatus(evse.CustomerId.ToString(), evse.ChargeBoxId, true, DefaultSetting.DefaultNullTime);

                            await UpdateOnlineRecords(evse.ChargeBoxId, true, evse.HeartbeatUpdatedOn, null);

                            logger.Debug(evse.ChargeBoxId + " Off-On Finished");
                        }

                    }
                    else
                    {

                        if (evse.Online)
                        {
                            //on -off  
                            logger.Debug(evse.ChargeBoxId + " On-Off Started");
                            var online_row =  await GetOnlineRecord(evse.ChargeBoxId);
                            if(online_row!=null)
                            {
                                await UpdateOnlineRecords(evse.ChargeBoxId, false, evse.HeartbeatUpdatedOn, online_row.Id);
                            }                         

                            logger.Debug(evse.ChargeBoxId + " On-Off Finished");
                        }

                    }
                }
            }
            catch (Exception ex)
            {
                logger.Debug("ERROR " + this.ToString() + ex.ToString());
            }



            logger.Debug(this.ToString() + " :Finished........");

        }


        private List<EVSECurrentStatus> GetEVSEs()
        {
            List<EVSECurrentStatus> result = new List<EVSECurrentStatus>();
            try
            {
                using (var dbConn = new SqlConnection(mainDBConnectString))
                {
                    dbConn.Open();
                    string sqlstring = "SELECT m.CustomerId,m.Id,m.ChargeBoxId,m.Online,m.HeartbeatUpdatedOn,MachineConfigurations.ConfigureSetting HeartbeatInterval"
                   + "  FROM [dbo].[Machine]  m  left join [dbo].[MachineConfigurations]  MachineConfigurations  on m.ChargeBoxId = MachineConfigurations.ChargeBoxId"
                   + " where MachineConfigurations.ConfigureName = 'HeartbeatInterval'  and MachineConfigurations.ConfigureSetting!=''";
                    result = dbConn.Query<EVSECurrentStatus>(sqlstring).ToList();
                }

            }
            catch (Exception ex)
            {
                logger.Error("Query Data Error " + ex.ToString());
            }


            return result;
        }

        private void UpdateEVSECurrentStatus(string customerId, string ChargeBoxId, bool turnOn, DateTime offlineTime)
        {
            try
            {
                string sqlString = string.Format("UPDATE [dbo].[Machine] SET Online=@Online {0} WHERE chargeBoxId=@chargeBoxId and customerId=@customerId", turnOn ? "" : " ,OfflineOn=@OfflineOn");
                using (var dbConn = new SqlConnection(mainDBConnectString))
                {
                    dbConn.Open();
                    var parameters = new DynamicParameters();
                    parameters.Add("@Online", turnOn, System.Data.DbType.Boolean);
                    parameters.Add("@chargeBoxId", ChargeBoxId, System.Data.DbType.String);
                    parameters.Add("@customerId", customerId, System.Data.DbType.String);
                    if (!turnOn)
                    {
                        parameters.Add("@OfflineOn", offlineTime, System.Data.DbType.DateTime);
                    }



                    dbConn.Execute(sqlString, parameters);
                }

            }
            catch (Exception ex)
            {
                logger.Error("Update Data Error " + ex.ToString());
            }
        }


        async private Task UpdateOnlineRecords(string chargeBoxId, bool online, DateTime hearbeatDt, Int64? rowId)
        {
            try
            {
                if (online)
                {
                    string sqlString = "INSERT INTO [dbo].[EVSEOnlineRecord] (\"ChargeBoxId\",\"OnlineTime\",\"OfflineTime\")" +
                       "VALUES( @ChargeBoxId,@OnlineTime,@OfflineTime); ";
                    using (var dbConn = new SqlConnection(onlineDBConnectString))
                    {
                        dbConn.Open();
                        var parameters = new DynamicParameters();
                        parameters.Add("@ChargeBoxId", chargeBoxId, System.Data.DbType.String);
                        parameters.Add("@OnlineTime", hearbeatDt, System.Data.DbType.DateTime);
                        parameters.Add("@OfflineTime", DefaultSetting.DefaultNullTime, System.Data.DbType.DateTime);

                        await dbConn.ExecuteAsync(sqlString, parameters);
                    }
                }
                else
                {
                    if (rowId.HasValue)
                    {
                        string sqlString = "UPDATE [dbo].[EVSEOnlineRecord] SET OfflineTime=@OfflineTime  WHERE Id=@Id";
                        using (var dbConn = new SqlConnection(onlineDBConnectString))
                        {
                            dbConn.Open();
                            var parameters = new DynamicParameters();
                            parameters.Add("@OfflineTime", hearbeatDt, System.Data.DbType.DateTime);
                            parameters.Add("@Id", rowId.Value, System.Data.DbType.Int64);

                            await dbConn.ExecuteAsync(sqlString, parameters);
                        }
                    }
                    else
                    {
                        logger.Error("Can't find update row id from " + chargeBoxId);
                    }

                }
            }
            catch (Exception ex)
            {
                logger.Error(string.Format("UpdateOnlineRecords Exception:{0}", ex.ToString()));
            }




        }



        async private Task<EVSEOnlineRecord> GetOnlineRecord(string chargeBoxId)
        {
            EVSEOnlineRecord lastrow = new EVSEOnlineRecord();
            try
            {
                string sqlString = string.Format("SELECT Id FROM [dbo].[EVSEOnlineRecord] WHERE chargeBoxId=@chargeBoxId Order by OnlineTime desc");
                using (var dbConn = new SqlConnection(onlineDBConnectString))
                {
                    dbConn.Open();
                    var parameters = new DynamicParameters();
                    parameters.Add("@chargeBoxId", chargeBoxId, System.Data.DbType.String);


                    var result = await dbConn.QueryAsync<EVSEOnlineRecord>(sqlString, parameters);
                    lastrow = result.FirstOrDefault();
                }

            }
            catch (Exception ex)
            {
                logger.Error("Query Data Error " + ex.ToString());
            }
            return lastrow;
        }     


        private bool IsOnlineNow(EVSECurrentStatus currentEVSE)
        {
            bool isOnline = false;
            int heartbeatInterval = 60;
            int.TryParse(currentEVSE.HeartbeatInterval, out heartbeatInterval);
            var checkTime = DateTime.UtcNow.AddSeconds(-(heartbeatInterval * 2));

            if (currentEVSE.HeartbeatUpdatedOn > checkTime)
            {
                isOnline = true;

            }

            return isOnline;

        }
    }
}
